/** +--------------------------------------------------------------------+
 *  | relay card v1.0 - firmware                                         |
 *  |   by mgr, 11.2007                                                  |
 *  |   last change: 2007-12-02                                          |
 *  |                                                                    |
 *  | This is the ATmega32 firmware for the relay card version 1.0.      |
 *  | For the circuit and further usage notes, have a look at the        |      
 *  | project homepage. You should program the high fuse byte to 0xcf    |
 *  | and the low byte to 0x3f.                                          |
 *  |                                                                    |
 *  | In order to compile this program, you will need a working AVR      | 
 *  | gcc cross compiler. You can also use the precompiled hex file      |
 *  | in this package if you do not want to make changes.                |
 *  |                                                                    |
 *  | For questions and suggestions: mailto michael@coremelt.net         |
 *  +--------------------------------------------------------------------+
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <stdint.h>

/* if you use an oscillator other than 8MHz, you'll need to change the value here */
#ifndef F_CPU
#  define F_CPU 8000000UL
#endif

#include <util/delay.h>

/* commonly used between firmware and software client */
#include "communication.h"

#define UART_BAUD_RATE 9600L
#define UART_UBRR_CALC(_BAUD, _FREQ) (uint16_t)((_FREQ)/((_BAUD)*16L)-1L)

/* some helper macros */
#define RELAY_STATUS()  (PORTD & 0xfe)
#define RELAY_NUM(n)    (1 << ((n)-1))
#define AUX_LED_OFF(n)  PORTA |= (1 << (n))
#define AUX_ALL_OFF()   PORTA |= 0x7e
#define AUX_LED_ON(n)   PORTA &= ~(1 << (n))
#define RXC_INT_ON()    UCSRB |= (1 << RXCIE)
#define RXC_INT_OFF()   UCSRB &= ~(1 << RXCIE)
#define USART_RX_ON()   UCSRB |= (1 << RXEN) 
#define USART_RX_OFF()  UCSRB &= ~(1 << RXEN)

/* RC5 stuff */
#define RC5TIME 	1.778e-3		// 1.778msec
#define PULSE_MIN	(uint8_t)(F_CPU / 512 * RC5TIME * 0.4 + 0.5)
#define PULSE_1_2	(uint8_t)(F_CPU / 512 * RC5TIME * 0.8 + 0.5)
#define PULSE_MAX	(uint8_t)(F_CPU / 512 * RC5TIME * 1.2 + 0.5)

/* status- and command codes */
#define RELAYOP_ON     0x01
#define RELAYOP_OFF    0x02
#define RELAYOP_SET    0x03
#define RELAYOP_TOGGLE 0x04

#define TIMER_TYPE_ON       0x01
#define TIMER_TYPE_OFF      0x02
#define TIMER_TYPE_CYCLIC   0x04
#define TIMER_UNIT_SECS     10
#define TIMER_DISP_INTERVAL 3
#define MODE_CHANGE_SECS    5

#define MODE_MANUAL_SETUP   0x01
#define MODE_DISPLAY_TIMING 0x02

/* function prototypes */
void initial_setup(void);
void display(uint8_t disp1, uint8_t disp2, uint8_t dec_point, uint8_t off);
void display_decimal(uint8_t value);
void relay_control(uint8_t mask, uint8_t operation);
inline uint8_t debounce(volatile uint8_t* port, uint8_t pin);
inline void sendc(uint8_t c);
inline uint8_t receivec();
void sendstr(char* str, uint8_t len);
uint8_t reverse_bitmask(uint8_t mask);

typedef struct
{
  uint8_t  active;
  uint8_t  type;
  uint32_t units;
  uint32_t reset_value;
} relay_timer;

/* global variables */
volatile relay_timer timers[6];
volatile uint8_t mode_change_secs=0, mode=MODE_MANUAL_SETUP;

volatile uint16_t rc5_tmp, rc5_data;
volatile uint8_t  rc5_bit, rc5_time;

/* This interrupt handler is the first output compare handler of timer 1 (16-bit).
 * It is configured to trigger with a frequency of one second exactly (see initial_setup()). */
ISR(TIMER1_COMPA_vect)
{
  static uint8_t secs=0, current_timer=0;
  uint8_t i, j;
  uint32_t value;
  
  /* update and evaluate all active timers */
  for (i=0; i<=5; i++)
  {
    if (timers[i].active)
    {
      switch (timers[i].type)
      {
        case TIMER_TYPE_ON:
	  if (--(timers[i].units) == 0)
	  {
	    relay_control((1 << i), RELAYOP_ON);
	    timers[i].active = 0;
	  }  
	  break;
	
	case TIMER_TYPE_OFF:
	  if (--(timers[i].units) == 0)
	  {
	    relay_control((1 << i), RELAYOP_OFF);
	    timers[i].active = 0;
	  }
	  break;
	
	case TIMER_TYPE_CYCLIC:
	  if (--(timers[i].units) == 0)
	  {
	    relay_control((1 << i), RELAYOP_TOGGLE);
	    
	    timers[i].units = timers[i].reset_value;
	  }
	  break;
	
	default:
	  break;
      }
    }
  }
  
  /* display information about the timers */
  if (mode_change_secs > 0)
    --mode_change_secs;
  
  if ((mode == MODE_DISPLAY_TIMING) || !mode_change_secs)
  {
    mode = MODE_DISPLAY_TIMING;
    AUX_ALL_OFF();

    if (!((++secs)%TIMER_DISP_INTERVAL))
    {
      secs = 0;
      
      for (j=(current_timer+1)%6; j!=current_timer; j = (j+1)%6)
      {
        if (timers[j].active)
        {
          current_timer = j;
	  break;
        }
      }
    }

    if (current_timer < 6 && timers[current_timer].active)
    { 
       AUX_LED_ON(6-current_timer);
       /* The timers hold their values in seconds. If we have values greater than one
	* minute, we'll display minutes, else we display seconds. If the value is greater
	* than 99 minutes, we just keep it at 99. */
       value = timers[current_timer].units;

       if (value <= 60)
       {
         display_decimal(value);
       } else
       {
         display_decimal((uint8_t)(value/60));
       }
    } else
    {  /* no active timer, switch the display off */
       display(0, 0, 0, 1);
    }
  } 
}

/* 8 bit timer. This timer is used to decode the information provided by the infra red
 * receiver. I've adapted this code from an example provided by Peter Danegger.
 * TODO: does not work yet */
ISR(TIMER0_OVF_vect)
{
#if 0
  uint16_t tmp = rc5_tmp;			// for faster access

  TCNT0 = -2;					// 2 * 256 = 512 cycle

  if (++rc5_time > PULSE_MAX)
  {			                        // count pulse time
    if(!(tmp & 0x4000) && tmp & 0x2000)	        // only if 14 bits received
      rc5_data = tmp;
    
    tmp = 0;
  }

  if ((rc5_bit ^ PINA) & (1 << PA0))
  {		                                // change detect
    rc5_bit = ~rc5_bit;				// 0x00 -> 0xFF -> 0x00

    if(rc5_time < PULSE_MIN)			// to short
      tmp = 0;

    if(!tmp || rc5_time > PULSE_1_2)
    {		                                // start or long pulse time
      if(!(tmp & 0x4000))			// not to many bits
        tmp <<= 1;				// shift
      
      if(!(rc5_bit & (1 << PA0)))		// inverted bit
        tmp |= 1;				// insert new bit
      
      rc5_time = 0;				// count next pulse time
    }
  }

  rc5_tmp = tmp;
#endif
}

/* This interrupt is called if a character was received by the UART. If we get
 * data, it is interpreted as a command. Commands have several more arguments to
 * it, these will be retrieved if required. */
ISR(USART_RXC_vect)
{
  uint8_t buf[COMMANDO_LENGTH];
  uint8_t i;
  
  /* disable this interrupt so we don't run into events that have already been
   * evaluated. We read several more bytes in this routine which would create
   * pending interrupts that are already served. */
  RXC_INT_OFF();

  /* Detect framing- and overrun errors. These bits are only valid until UDR is read. */
  if ((UCSRA & (1 << FE)) || (UCSRA & (1 << DOR)))
  {
    sendc(RESPONSE_TRANSMISSION_ERROR);
    buf[0] = UDR; // read UDR anyway so the interrupt gets cleared
    RXC_INT_ON();
    return;
  }

  /* The first byte of the transmission is the command byte.
   * We must read UDR, otherwhise new interrupts would be generated until this is done.*/
  buf[0] = UDR;

  /* Fetch the argument(s) of the command. */
  buf[1] = receivec();
  buf[2] = receivec();
  buf[3] = receivec();

  /* evaluate the commands */
  switch (buf[0])
  {
    case COMMAND_RELAY_ON:
    case COMMAND_RELAY_OFF:
    case COMMAND_RELAY_TOGGLE:
    case COMMAND_RELAY_SET:
     
      if (buf[1] > ((1 << 7)-1))
      {
	_delay_ms(5);
	sendc(RESPONSE_INVALID_ARGUMENT);
        RXC_INT_ON();
	return;
      }
      
      if (buf[0] == COMMAND_RELAY_ON)
        relay_control(buf[1], RELAYOP_ON);
      else if (buf[0] == COMMAND_RELAY_OFF)
        relay_control(buf[1], RELAYOP_OFF);
      else if (buf[0] == COMMAND_RELAY_TOGGLE)
        relay_control(buf[1], RELAYOP_TOGGLE);
      else
       relay_control(buf[1], RELAYOP_SET);

      break;

    case COMMAND_GET_STATUS:
      _delay_ms(5);
      sendc(reverse_bitmask((PORTD & 0xfc) >> 2));
      RXC_INT_ON();
      return;

    case COMMAND_RELAY_TIME_ON:
      timers[buf[1]].active = 1;
      timers[buf[1]].type = TIMER_TYPE_ON;
      timers[buf[1]].units = ((buf[3] << 8) | buf[2])*TIMER_UNIT_SECS;

      relay_control((1 << buf[1]), RELAYOP_OFF);   
      break;
    
    case COMMAND_RELAY_TIME_OFF:
      timers[buf[1]].active = 1;
      timers[buf[1]].type = TIMER_TYPE_OFF;
      timers[buf[1]].units = ((buf[3] << 8) | buf[2])*TIMER_UNIT_SECS;

      relay_control((1 << buf[1]), RELAYOP_ON);
      break;
    case COMMAND_RELAY_TIME_CYCLIC:
      timers[buf[1]].active = 1;
      timers[buf[1]].type = TIMER_TYPE_CYCLIC;
      timers[buf[1]].units = ((buf[3] << 8) | buf[2])*TIMER_UNIT_SECS;
      timers[buf[1]].reset_value = timers[buf[1]].units; 

      relay_control((1 << buf[1]), RELAYOP_ON);
      break;

    case COMMAND_DEL_TIMERS:
      for (i=0; i<6; i++)
      {
        timers[i].active = 0;
      }
      break;

    default:
      _delay_ms(5);
      sendc(RESPONSE_INVALID_COMMAND);
      RXC_INT_ON();
      return;
  }
  
  _delay_ms(5); 
  sendc(RESPONSE_OK);
  
  /* enable the interrupt again */
  RXC_INT_ON();
}

int main(void)
{
    uint8_t selected_relay=0, i=0;

    /* in this application, the watchdog will not be used */
    wdt_disable();
    
    /* setup the controller */
    initial_setup();
    
    /* initialize the timers */
    for (i=0; i<6; i++)
    {
      timers[i].active = 0;
    }
    
    /* program main loop
     * some of the less important things are handled here ;) */
    while (1)
    {
      // ========================= manual relay control
      if (debounce(&PINB, PB4))
      {
	mode = MODE_MANUAL_SETUP;
	mode_change_secs = MODE_CHANGE_SECS;
	display(0, 0, 0, 1);
	AUX_ALL_OFF();
        
	if (++selected_relay > 6)
	{
	  selected_relay = 0;
	}
        
	if (selected_relay)
	  AUX_LED_ON(7 - selected_relay);
      }
      
      if ((mode == MODE_MANUAL_SETUP) && selected_relay && debounce(&PINA, PA7))
      {
        relay_control((1 << (selected_relay-1)), RELAYOP_TOGGLE);
	timers[selected_relay-1].active = 0;
      }
      // ========================= manual relay control end
    } 
    
    return 0;
}

/* Display control. Pass the values from 1-9 (decimal) to disp1 and disp2. To
 * switch off the display completely, off must be true. If dec_point is true,
 * the decimal point on DISP1 will be switched on, but it is not possible to
 * use the decimal point if the displays are switched off.
 */
void display(uint8_t disp1, uint8_t disp2, uint8_t dec_point, uint8_t off)
{
  if (disp1 > 9)
    disp1 = 9;
  
  if (disp2 > 9)
    disp2 = 9;

  if (off)
  { /* to switch the display off, PB0 and PB1 must be set to LOW */
    PORTB &= ~((1 << PB0) | (1 << PB1));
    return;
  } else
  {
    PORTB |= ((1 << PB0) | (1 << PB1));
  }
  
  /* the decimal pin is controlled directly by PB2 */
  if (dec_point && !off)
  {
    PORTB &= ~(1 << PB2);
  } else
  {
    PORTB |= (1 << PB2);
  }
  
  /* Thanks to the BCD-decoders used, we can output the desired values directly
   * to PORTC. The lower four bits are for DISP1, the higher four are for DISP2. */
  PORTC = disp1 | (disp2 << 4);
}

/* display a decimal value from 0-99 */
void display_decimal(uint8_t value)
{
  uint8_t lower, higher;

  if (value > 99)
    value = 99;
 
  lower = value%10;
  higher = (value-lower)/10;

  display(higher, lower, 0 ,0);
}

uint8_t reverse_bitmask(uint8_t mask)
{
  uint8_t i;
  uint8_t reverse_mask=0;

  for (i=0; i<=5; i++)
  {
    if (mask & (1 << i))
    {
      reverse_mask |= (1 << (5-i));
    }
  }

  return reverse_mask;
}

/* This function controls the relays. Notice that the bit mask "mask" addresses
 * the relays so that the LSB of the mask switches the right most relay on the board.
 */
void relay_control(uint8_t mask, uint8_t operation)
{
  uint8_t reverse_mask=0;

  if (mask > 0x3f)
  {
    mask = 0x3f;
  }
  
  /* reverse the bit mask */
  reverse_mask = reverse_bitmask(mask);

  switch (operation)
  {
    case RELAYOP_ON:
      PORTD |= (reverse_mask << 2); 
      break;
    
    case RELAYOP_OFF:
      PORTD &= ~(reverse_mask << 2);
      break;
    
    case RELAYOP_TOGGLE:
      PORTD ^= (reverse_mask << 2);
      break;

    case RELAYOP_SET:
      PORTD = (reverse_mask << 2);

    default:
      break;
  }
}

/* simple switch debouncing function - works quite well ;) */
inline uint8_t debounce(volatile uint8_t* port, uint8_t pin)
{
  if (!(*port & (1 << pin)))
  {
      _delay_ms(20);

      if (*port & (1 << pin))
      {
        return 1;
      }
  }

  return 0;
}

/* send a character over the USART */
inline void sendc(uint8_t c)
{
  loop_until_bit_is_set(UCSRA, UDRE);
  UDR = c;
}

/* receive a character over the USART */
inline uint8_t receivec()
{
  loop_until_bit_is_set(UCSRA, RXC);
  return UDR;
}

/* Send a string over the USART. No further tests are performed. */
void sendstr(char* str, uint8_t len)
{
  int i;

  for (i=0; i<len; i++)
  {
    sendc((uint8_t)str[i]);
    if (str[i] == '\0')
      break;
  }
}

void initial_setup(void)
{
 /* Here, the port setup is done. For more clarity I recommend having a look at the
  * schematic of the board, which can (also) be obtained from the project homepage.
  * For inputs, the internal pullups will be enabled.
  *
  * Port A.1-A.6 is used for the auxiliary LEDS 1-7, A.0 is the input for the TSOP and
  * A.7 is used by SW2. All outputs are initially set to high (LEDs off).
  */
  DDRA  =  (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4) | (1 << PA5) | (1 << PA6);
  PORTA =  (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4) | (1 << PA5) | (1 << PA6) 
	   | (1 << PA7);

 /* Port B is used for miscellaneous stuff.
  * B.0 and B.1 are used for the blank input lines of DISP1 and DISP2,
  * B.2 controlls the decimal point of DISP1.
  *
  * B.4 is used for SW1.
  * B.7 is the TSOP output line (input on the controller).
  *
  * Upon startup, the display gets switched off (B.0 and B.1 LOW). B.2 must be HIGH in
  * order to switch the dot off.
  */
  DDRB = (1 << PB0) | (1 << PB1) | (1 << PB2);
  PORTB = (1 << PB2) | (1 << PB4);

 /* C.0-C.3 are the BCD-Code lines for DISP1, C.4-C.7 for DISP2, respectively, so they
  * will be configured as outputs alltogether.
  */
  DDRC = 0xff;
  PORTC = 0x00; 
  
  /* The relays R1-R6 are attached to D.2-D.7 of Port D. D.0 and D.1 are used as RXD and
   * TXD lines for the USB UART and not used further. All outputs are set to low, meaning
   * relay OFF.
   */
  DDRD = (1 << PD2) | (1 << PD3) | (1 << PD4)  | (1 << PD5) | (1 << PD6) |  (1 << PD7);
  PORTD = 0x00;

  /* Initializes the USART */
  UBRRH = (uint8_t)(UART_UBRR_CALC(UART_BAUD_RATE, F_CPU) >> 8);
  UBRRL = (uint8_t)(UART_UBRR_CALC(UART_BAUD_RATE, F_CPU));

  /* frame format: 8 data bits, 2 stop bits, no parity */
  UCSRA = 0x00; 
  UCSRB = (1 << TXEN) | (1 << RXEN) | (1 << RXCIE); // enable RX, TX and interrupts
  UCSRC = (1 << URSEL) | (3 << UCSZ0) | (1 << USBS);

  /* configure the 8-Bit-Timer */
  TCCR0 |= (1 << CS01); // F_CPU/8
  TIMSK |= (1 << TOIE0);
  
  /* configure the 16 bit timer to a one second interval */
  TCCR1B |= (1 << CS12) | (1 << FOC1A); // F_CPU/256
  OCR1A = 0x7a12;                       // compare value: 8000000/256 = 31250 = 0x7a12
  TIMSK |= (1 << OCIE1A);               // enable the timer compare match interrupt 

  /* enable interrupts */
  sei();
}
